package springBugs.configInheritance;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Objects;
import java.util.stream.Collectors;

import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.AnnotationConfigApplicationContext;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;

public class SpringConfigInheritance
{

	@Configuration
	public static class ParentConfig
	{
		@Bean
		public String bean()
		{
			return "parent";
		}
		
		@Bean
		public String configClass()
		{
			return getClass().getName();
		}
	}
	
	@Configuration
	public static class ChildConfig extends ParentConfig
	{
		@Bean("renamed-bean")
		@Override
		public String bean()
		{
			return "child";
		}
	}
	
	public static void main(String[] args) throws IllegalAccessException, IllegalArgumentException, InvocationTargetException
	{
	
		System.out.println("Given normal inheritance rules and the rules for annotations on methods the expected output would be:");
	
		ChildConfig childConfig = new ChildConfig();
		for ( Method method : ChildConfig.class.getMethods() )
		{
			if ( method.isAnnotationPresent(Bean.class) )
			{
				String[] names = method.getAnnotation(Bean.class).value();
				Object   value = method.invoke(childConfig);
				System.out.println("@Bean method: " + method.getName() + 
								   "\n\tBean names: " + Arrays.stream(names).collect(Collectors.joining(", ")) +
								   "\n\tBean value: " + Objects.toString(value));
			}
		}
		
		System.out.println();
		System.out.println("What spring gives us is actually: ");
		
		// Parent with one child config
		try ( AnnotationConfigApplicationContext context = new AnnotationConfigApplicationContext() )
		{
			context.register(ChildConfig.class);
			context.refresh();
			
			for ( String name : context.getBeanDefinitionNames() )
			{
				BeanDefinition definition = context.getBeanDefinition(name);
				if ( definition.getRole() == BeanDefinition.ROLE_APPLICATION && definition.getFactoryMethodName() != null)
				{
					Object value = context.getBean(name);
					
					System.out.println("@Bean method: " + definition.getFactoryMethodName() + 
								   "\n\tBean names: " + name +", " + Arrays.stream(context.getAliases(name)).collect(Collectors.joining(", ")) +
								   "\n\tBean value: " + Objects.toString(value));				
				}
			}
			
			System.out.println();
			System.out.println("Let's see if this is consistent with an inheritence == inclusion model by running some tests:");

			// Why does this bean exist?  We only included ChildConfig.
			// Okay, so inheritance == inclusion?  Maybe this make sense because of parallels to XML config?
			boolean hasParentBean = context.containsBean("bean");
			System.out.printf("parent bean existence: %s -> inheritance %s inclusion\n", hasParentBean, hasParentBean ?"==":"!=");
			
			if ( hasParentBean )
			{
				// Let's double check that it isn't just an alias now, which would be consistent with inclusion
				boolean isAlias = context.isAlias("bean");
				System.out.printf("parent bean is alias test: %s -> inheritance %s inclusion\n", isAlias, isAlias?"!=":"==");
				
				// Let's double check that it isn't just an alias now, which would be consistent with inclusion
				isAlias = context.isAlias("renamed-bean");
				System.out.printf("child bean is alias test: %s -> inheritance %s inclusion\n", isAlias, isAlias?"!=":"==");
				
				// Alright, so if inheritance == inclusion, then the value should be "parent" 
				String parentBean = context.getBean("bean", String.class);
				boolean isExpectedValueViaInclusion = parentBean.equals("parent");
				System.out.printf("parent bean value == \"parent\" test: %s -> inheritance %s inclusion\n", isExpectedValueViaInclusion, isExpectedValueViaInclusion?"==":"!=");
				
				// Alright, so if inheritance == inclusion, then the configClass value should be "ParentConfig"???
				String configClass = context.getBean("configClass", String.class);
				boolean isExpectedConfigClassViaInclusion = configClass.equals("springBugs.configInheritance.SpringConfigInheritance$ParentConfig");
				System.out.printf("parent configClass value == \"springBugs.configInheritance.SpringConfigInheritance$ParentConfig\" test: %s -> inheritance %s inclusion\n", isExpectedConfigClassViaInclusion, isExpectedConfigClassViaInclusion?"==":"!=");
			}
		}
	}
}
